#cython: language_level=3

from obitools3.dms.capi.apat cimport SeqPtr


cdef class One_primer_search_result:
    cdef SeqPtr _pointer
    cdef int pattern_ref
    cdef int hit_count
    cdef inline SeqPtr pointer(self)
    @staticmethod
    cdef new(SeqPtr apat_seq_p, int pattern_ref, int hit_count)
    cpdef first_encountered(self)
    

cdef class Primer_search:
    cdef SeqPtr apat_seq_p
    cdef list direct_primers
    cdef list revcomp_primers
    cpdef One_primer_search_result search_one_primer(self, bytes sequence, 
                                                     int primer_pair_index, int primer_index, 
                                                     bint reverse_comp=*, 
                                                     bint same_sequence=*, 
                                                     int pattern_ref=*,
                                                     int begin=*)
    cpdef free(self)
    




import glob
import os
import sys
import re
import subprocess

from distutils import log
#from distutils.core import setup
from setuptools import setup    # to work with pip

from distutils.core import Extension
from distutils.sysconfig import get_python_lib

import os.path
from distutils import log
from distutils.extension import Extension

from distutils.dist import Distribution as ori_Distribution

from python.obitools3.version import version


class Distribution(ori_Distribution):
    
    def __init__(self,attrs=None):
        self.cobitools3=attrs['cobitools3']
        
        ori_Distribution.__init__(self, attrs)
        
        self.global_options.insert(0,('cobitools3', None, "install location of the C library"
                                     ))

from distutils.command.build import build as build_ori
from setuptools.command.bdist_egg import bdist_egg as bdist_egg_ori
from distutils.core import Command


class build_clib(Command):
    user_options=[]
    
    def initialize_options(self):
        self.clib_dir=self.distribution.cobitools3
    
    def finalize_options(self):
        if self.clib_dir is None:
            self.clib_dir=get_python_lib()
    
    def run(self):
        log.info("Build the build/cobject directory")
        try:
            os.mkdir("build")
        except OSError:
            pass
        try:
            os.mkdir("build/cobject")
        except OSError:
            pass
        
        oldwd = os.getcwd()
        os.chdir("build/cobject")
        install_clibdir_option="-DPYTHONLIB:STRING='%s'" % self.clib_dir
        log.info("Run CMake")
        subprocess.call(['cmake', install_clibdir_option, '../../src'])
        log.info("Compile the shared C library")
        subprocess.call(['make','install'])   # temporary fix but should be in src
        os.chdir(oldwd)


class build(build_ori):
    def run(self):
        self.run_command("build_clib")
        build_ori.run(self)


class bdist_egg(bdist_egg_ori):
    def run(self):
        self.run_command('build_clib')
        bdist_egg_ori.run(self)


sys.path.append(os.path.abspath("python"))


def findPackage(root,base=None):
    modules=[]
    if base is None:
        base=[]
    for module in (os.path.basename(os.path.dirname(x)) 
                   for x in glob.glob(os.path.join(root,'*','__init__.py'))):
        modules.append('.'.join(base+[module]))
        modules.extend(findPackage(os.path.join(root,module),base+[module]))
    return modules
  

PACKAGE     = "OBITools3"
VERSION     = version
AUTHOR      = 'Celine Mercier'
EMAIL       = 'celine.mercier@metabarcoding.org'
URL         = "https://metabarcoding.org/obitools3"
PLATFORMS   = "posix"
LICENSE     = "CeCILL-V2"
DESCRIPTION = "A package for the management of analyses and data in DNA metabarcoding."
PYTHONMIN   = '3.5'

SRC       = 'python'
CSRC      = 'src'

REQUIRES  = ['Cython>=0.24',
             'Sphinx>=1.2.0',
             'ipython>=3.0.0',
             'breathe>=4.0.0'
            ]

os.environ['CFLAGS'] = '-O3 -w -I "src" -I "src/libecoPCR" -I "src/libjson"'


from Cython.Build import cythonize

cython_src  = [x for x in glob.iglob('python/obitools3/**/*.pyx', 
                                     recursive=True
                                    )
              ]
      

cython_ext  = [Extension('.'.join([os.path.dirname(x).replace("python/",""),
                                   os.path.splitext(os.path.basename(x))[0]]).replace('/','.'),
                         [x],
                         library_dirs=[get_python_lib()],
                         include_dirs=["src","src/libecoPCR","src/libjson"],
                         libraries=["cobitools3"],
                         runtime_library_dirs=[get_python_lib()],
                         extra_compile_args=['-msse2',
                                          '-w',
                                          '-fPIC'
                                         ],
                         extra_link_args=["-Wl,-rpath,"+get_python_lib(), 
                                          "-L"+get_python_lib()
                                         ]
                        )
                for x in cython_src
              ]              

xx = cythonize(cython_ext,
               language_level=3,
               annotate=True,
               build_dir="build")

classifiers=['Development Status :: 4 - Beta',
             'Environment :: Console',
             'Intended Audience :: Science/Research',
             'License :: Other/Proprietary License',
             'Operating System :: Unix',
             'Programming Language :: Python :: 3',
             'Programming Language :: C',
             'Topic :: Scientific/Engineering :: Bio-Informatics',
             'Topic :: Utilities',
             ]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name=PACKAGE,
      description=DESCRIPTION,
      long_description=long_description,
      long_description_content_type="text/markdown",
      classifiers=classifiers,
      version=VERSION,
      author=AUTHOR,
      author_email=EMAIL,
      platforms=PLATFORMS,
      license=LICENSE,
      url=URL,
      ext_modules=xx,
      distclass=Distribution,
      cmdclass={'build': build,
                'bdist_egg': bdist_egg,
                'build_clib': build_clib},
      cobitools3=get_python_lib(),
      packages = findPackage('python'),
      package_dir = {"" : "python"},
      scripts = ['scripts/obi']
)




#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # @UnresolvedImport
from obitools3.dms import DMS
from obitools3.dms.view.view cimport View, Line_selection
from obitools3.uri.decode import open_uri
from obitools3.apps.optiongroups import addMinimalInputOption, addTaxonomyOption, addMinimalOutputOption, addNoProgressBarOption
from obitools3.dms.view import RollbackException
from obitools3.dms.column.column cimport Column
from functools import reduce
from obitools3.apps.config import logger
from obitools3.utils cimport tobytes, str2bytes, tostr
from io import BufferedWriter
from obitools3.dms.capi.obiview cimport NUC_SEQUENCE_COLUMN, \
                                        ID_COLUMN, \
                                        DEFINITION_COLUMN, \
                                        QUALITY_COLUMN, \
                                        COUNT_COLUMN, \
                                        TAXID_COLUMN
from obitools3.dms.capi.obitypes cimport OBI_INT
from obitools3.dms.capi.obitaxonomy cimport MIN_LOCAL_TAXID
import time
import math
import sys

from cpython.exc cimport PyErr_CheckSignals


__title__="Add taxa with a new generated taxid to an NCBI taxonomy database"



def addOptions(parser):

    addMinimalInputOption(parser)
    addTaxonomyOption(parser)
    addMinimalOutputOption(parser)
    addNoProgressBarOption(parser)

    group=parser.add_argument_group('obi taxonomy specific options')

    group.add_argument('-n', '--taxon-name-tag',
                       action="store",
                       dest="taxonomy:taxon_name_tag",
                       metavar="<SCIENTIFIC_NAME_TAG>",
                       default=b"SCIENTIFIC_NAME",
                       help="Name of the tag giving the scientific name of the taxon "
                            "(default: 'SCIENTIFIC_NAME').")

#    group.add_argument('-g', '--try-genus-match',
#                     action="store_true", dest="taxonomy:try_genus_match",
#                     default=False,
#                     help="Try matching the first word of <SCIENTIFIC_NAME_TAG> when can't find corresponding taxid for a taxon. "
#                          "If there is a match it is added in the 'parent_taxid' tag. (Can be used by 'obi taxonomy' to add the taxon under that taxid).")

    group.add_argument('-a', '--restricting-ancestor',
                       action="store",
                       dest="taxonomy:restricting_ancestor",
                       metavar="<RESTRICTING_ANCESTOR>",
                       default=None,
                       help="Enables to restrict the addition of taxids under an ancestor specified by its taxid.")

    group.add_argument('-t', '--taxid-tag',
                       action="store",
                       dest="taxonomy:taxid_tag",
                       metavar="<TAXID_TAG>",
                       default=b"TAXID",
                       help="Name of the tag to store the new taxid "
                            "(default: 'TAXID').")

    group.add_argument('-l', '--log-file',
                       action="store",
                       dest="taxonomy:log_file",
                       metavar="<LOG_FILE>",
                       default='',
                       help="Path to a log file to write informations about added taxids.")


def run(config):

    DMS.obi_atexit()

    logger("info", "obi taxonomy")

    # Open the input
    input = open_uri(config['obi']['inputURI'])
    if input is None:
        raise Exception("Could not read input view")
    i_dms = input[0]
    i_view = input[1]
    i_view_name = input[1].name

    # Open the output: only the DMS, as the output view is going to be created by cloning the input view
    # (could eventually be done via an open_uri() argument)
    output = open_uri(config['obi']['outputURI'],
                      input=False,
                      dms_only=True)
    if output is None:
        raise Exception("Could not create output view")
    o_dms = output[0]
    output_0 = output[0]
    o_view_name = output[1]

    # stdout output: create temporary view
    if type(output_0)==BufferedWriter:
        o_dms = i_dms
        i=0
        o_view_name = b"temp"
        while o_view_name in i_dms: # Making sure view name is unique in output DMS
            o_view_name = o_view_name+b"_"+str2bytes(str(i))
            i+=1
        imported_view_name = o_view_name

    # If the input and output DMS are not the same, import the input view in the output DMS before cloning it to modify it
    # (could be the other way around: clone and modify in the input DMS then import the new view in the output DMS)
    if i_dms != o_dms:
        imported_view_name = i_view_name
        i=0
        while imported_view_name in o_dms:  # Making sure view name is unique in output DMS
            imported_view_name = i_view_name+b"_"+str2bytes(str(i))
            i+=1
        View.import_view(i_dms.full_path[:-7], o_dms.full_path[:-7], i_view_name, imported_view_name)
        i_view = o_dms[imported_view_name]

    # Clone output view from input view
    o_view = i_view.clone(o_view_name)
    if o_view is None:
        raise Exception("Couldn't create output view")
    i_view.close()

    # Open taxonomy
    taxo_uri = open_uri(config['obi']['taxoURI'])
    if taxo_uri is None or taxo_uri[2] == bytes:
        raise Exception("Couldn't open taxonomy")
    taxo = taxo_uri[1]

    # Initialize the progress bar
    if config['obi']['noprogressbar'] == False:
        pb = ProgressBar(len(o_view), config)
    else:
        pb = None

    try:
        if config['taxonomy']['log_file']:
            logfile = open(config['taxonomy']['log_file'], 'w')
        else:
            logfile = sys.stdout
        if 'restricting_ancestor' in config['taxonomy']:
            res_anc = int(config['taxonomy']['restricting_ancestor'])
        else:
            res_anc = None
        taxid_column_name = config['taxonomy']['taxid_tag']
        parent_taxid_column_name = "PARENT_TAXID" # TODO macro
        taxon_name_column_name = config['taxonomy']['taxon_name_tag']
        taxid_column = Column.new_column(o_view, taxid_column_name, OBI_INT)
        if parent_taxid_column_name in o_view:
            parent_taxid_column = o_view[parent_taxid_column_name]
        else:
            parent_taxid_column = None
        #parent_taxid_column = Column.new_column(o_view, parent_taxid_column_name, OBI_INT)
        taxon_name_column = o_view[taxon_name_column_name]

        for i in range(len(o_view)):
            PyErr_CheckSignals()
            #if pb is not None:
            #    pb(i)
            taxon_name = taxon_name_column[i]
            taxon = taxo.get_taxon_by_name(taxon_name, res_anc)
            if taxon is not None:
                taxid_column[i] = taxon.taxid
                if logfile:
                    print(f"Found taxon '{tostr(taxon_name)}' already existing with taxid {taxid_column[i]}", file=logfile)
            else: # try finding genus or other parent taxon from the first word
                #print(i, o_view[i].id)
                if parent_taxid_column is not None and parent_taxid_column[i] is not None:
                    taxid_column[i] = taxo.add_taxon(taxon_name, 'species', parent_taxid_column[i])
                    if logfile:
                        print(f"Adding taxon '{tostr(taxon_name)}' under provided parent {parent_taxid_column[i]} with taxid {taxid_column[i]}", file=logfile)
                else:
                    taxon_name_sp = taxon_name.split(b" ")
                    taxon = taxo.get_taxon_by_name(taxon_name_sp[0], res_anc)
                    if taxon is not None:
                        parent_taxid_column[i] = taxon.taxid
                        taxid_column[i] = taxo.add_taxon(taxon_name, 'species', taxon.taxid)
                        if logfile:
                            print(f"Adding taxon '{tostr(taxon_name)}' under '{tostr(taxon.name)}' ({taxon.taxid}) with taxid {taxid_column[i]}", file=logfile)
                    else:
                        taxid_column[i] = taxo.add_taxon(taxon_name, 'species', res_anc)
                        if logfile:
                            print(f"Adding taxon '{tostr(taxon_name)}' under provided restricting ancestor {res_anc} with taxid {taxid_column[i]}", file=logfile)

        taxo.write(taxo.name, update=True)
        
    except Exception, e:
        raise RollbackException("obi taxonomy error, rollbacking view: "+str(e), o_view)

    #if pb is not None:
    #    pb(i, force=True)
    #    print("", file=sys.stderr)

    #logger("info", "\nTaxa already in the taxonomy: "+str(found_count)+"/"+str(len(o_view))+" ("+str(round(found_count*100.0/len(o_view), 2))+"%)")
    #logger("info", "\nParent taxids found: "+str(parent_found_count)+"/"+str(len(o_view))+" ("+str(round(parent_found_count*100.0/len(o_view), 2))+"%)")
    #logger("info", "\nTaxids not found: "+str(not_found_count)+"/"+str(len(o_view))+" ("+str(round(not_found_count*100.0/len(o_view), 2))+"%)")

    # Save command config in View and DMS comments
    command_line = " ".join(sys.argv[1:])
    input_dms_name=[input[0].name]
    input_view_name=[i_view_name]
    if 'taxoURI' in config['obi'] and config['obi']['taxoURI'] is not None:
        input_dms_name.append(config['obi']['taxoURI'].split("/")[-3])
        input_view_name.append("taxonomy/"+config['obi']['taxoURI'].split("/")[-1])
    o_view.write_config(config, "taxonomy", command_line, input_dms_name=input_dms_name, input_view_name=input_view_name)
    o_dms.record_command_line(command_line)

    #print("\n\nOutput view:\n````````````", file=sys.stderr)
    #print(repr(o_view), file=sys.stderr)

    # stdout output: write to buffer
    if type(output_0)==BufferedWriter:
        logger("info", "Printing to output...")
        o_view.print_to_output(output_0, noprogressbar=config['obi']['noprogressbar'])
        o_view.close()

    # If the input and the output DMS are different or if stdout output, delete the temporary imported view used to create the final view
    if i_dms != o_dms or type(output_0)==BufferedWriter:
        View.delete_view(o_dms, imported_view_name)
        o_dms.close(force=True)
    i_dms.close(force=True)

    logger("info", "Done.")



#cython: language_level=3

from obitools3.dms.capi.kmer_similarity cimport kmer_similarity, Obi_ali_p
from libc.stdint  cimport int32_t, uint8_t
from obitools3.dms.capi.obiview cimport Obiview_p
from obitools3.dms.capi.obidmscolumn cimport OBIDMS_column_p


cdef class Ali_shifted:
    cdef Obi_ali_p _pointer
    cdef inline Obi_ali_p pointer(self)
    cpdef free(self)
    @staticmethod
    cdef Ali_shifted new_ali(Obi_ali_p ali_p)


cdef class Kmer_similarity:
    cdef uint8_t         kmer_size
    cdef int32_t*        kmer_pos_array_p
    cdef int32_t         kmer_pos_array_height_a[1]
    cdef int32_t*        shift_array_p
    cdef int32_t         shift_array_height_a[1]
    cdef int32_t*        shift_count_array_p
    cdef int32_t         shift_count_array_height_a[1]
    cdef Obiview_p       view1_p
    cdef OBIDMS_column_p column1_p
    cdef OBIDMS_column_p qual_col1_p
    cdef Obiview_p       view2_p
    cdef OBIDMS_column_p column2_p
    cdef OBIDMS_column_p qual_col2_p
    cdef OBIDMS_column_p reversed_col_p
    cdef bint            build_consensus
    cpdef free(self)


#cython: language_level=3

'''
Created on 6 Nov. 2009

@author: coissac
'''

from ._dynamic cimport * 
from ._rassemble cimport RightDirectAssemble
from copy import deepcopy
import array


cdef class QSolexaRightDirectAssemble(RightDirectAssemble):             
 
     cdef double*       hError
     cdef double*       vError
     cdef object        a

     def __init__(self,match=4,mismatch=-4,opengap=-8,extgap=-2):  
         """
         Rapport entre score de match et mismatch:
             si mismatch = - match / 3
             alors quand scrore temps vers 0 et qu'il est impossible de decider
             pas de penalisation (s'=0) 
             si mismatch < - match / 3 la non decidabilite est penalisee.
         """           
         RightDirectAssemble.__init__(self,match,mismatch,opengap,extgap)

 
     cdef double matchScore(self,int h, int v):             
         cdef double score
         cdef double smatch
         cdef double smismatch
         cdef double hok=1-self.hError[h-1]             
         cdef double vok=1-self.vError[v-1]
         score=iupacPartialMatch(self.hSeq.sequence[h-1],self.vSeq.sequence[v-1])
         smatch=((4*hok*vok-hok-vok)*(self._match-self._mismatch)+self._match+2*self._mismatch)/3             
         smismatch=((hok+vok-4*hok*vok)*(self._match-self._mismatch)+2*self._match+7*self._mismatch)/9
         return smatch * score + smismatch * (1. - score)

     property seqA:
             def __get__(self):             
                 return self.horizontalSeq

             def __set__(self, seq):
                 cdef object oaddresse,olength
                 assert hasattr(seq, "quality"),"You must use sequence with quality indices"             
                 self.sequenceChanged=True
                 self.horizontalSeq=seq
                 self.hSeq=allocateSequence(self.horizontalSeq,self.hSeq)
                 (oaddress,olength)=seq.quality_array.buffer_info()             
                 self.hError=<double*><unsigned long int>oaddress

     property seqB:
             def __get__(self):             
                 return self.verticalSeq
 
             def __set__(self, seq):
                 cdef object oaddresse,olength
                 assert hasattr(seq, "quality"),"You must use sequence with quality indices"             
                 self.sequenceChanged=True
                 self.verticalSeq=seq
                 self.vSeq=allocateSequence(self.verticalSeq,self.vSeq)
                 (oaddress,olength)=seq.quality_array.buffer_info()             
                 self.vError=<double*><unsigned long int>oaddress


cdef class QSolexaRightReverseAssemble(QSolexaRightDirectAssemble):             
 
     cdef double matchScore(self,int h, int v):
         cdef double score
         cdef double smatch
         cdef double smismatch
         cdef double hok=1-self.hError[h-1]  
         cdef double vok=1-self.vError[self.vSeq.length - v]
         score=iupacPartialMatch(self.hSeq.sequence[h-1],self.vSeq.sequence[v-1])
         smatch=((4*hok*vok-hok-vok)*(self._match-self._mismatch)+self._match+2*self._mismatch)/3
         smismatch=((hok+vok-4*hok*vok)*(self._match-self._mismatch)+2*self._match+7*self._mismatch)/9       
         return smatch * score + smismatch * (1. - score)


     property seqB:
             def __get__(self):             
                 return self.verticalSeq.wrapped

             def __set__(self, seq):
                 cdef object oaddresse,olength
                 assert hasattr(seq, "quality"),"You must use sequence with quality indices"             
                 self.sequenceChanged=True
                 self.verticalSeq=seq.reverse_complement
                 self.vSeq=allocateSequence(self.verticalSeq,self.vSeq)
                 self.a = array.array('d', list(seq.quality_array))
                 (oaddress,olength) = self.a.buffer_info()
                 self.vError = <double*><unsigned long int>oaddress



#cython: language_level=3

cdef class TabFormat:
    cdef bint header
    cdef bint first_line
    cdef bytes NAString
    cdef set   tags
    cdef bytes sep
    cdef bint NAIntTo0
    cdef bint metabaR
    cdef bint ngsfilter



#cython: language_level=3

from obitools3.dms.capi.obierrno cimport obi_errno

from ..column cimport register_column_class

from ...view.view cimport View

from ..column cimport Column

from obitools3.utils cimport tobytes, bytes2str, \
                             obi_errno_to_exception

from ...capi.obiview cimport obi_get_qual_char_with_elt_name_and_col_p_in_view, \
                             obi_get_qual_char_with_elt_idx_and_col_p_in_view, \
                             obi_set_qual_char_with_elt_name_and_col_p_in_view, \
                             obi_set_qual_char_with_elt_idx_and_col_p_in_view, \
                             obi_get_qual_int_with_elt_name_and_col_p_in_view, \
                             obi_get_qual_int_with_elt_idx_and_col_p_in_view, \
                             obi_set_qual_int_with_elt_name_and_col_p_in_view, \
                             obi_set_qual_int_with_elt_idx_and_col_p_in_view, \
                             Obiview_p

from ...capi.obidmscolumn cimport OBIDMS_column_p

from ...capi.obitypes     cimport OBI_QUAL, OBIQual_char_NA, OBIQual_int_NA, const_char_p

from libc.stdlib  cimport malloc, free
from libc.stdint cimport uint8_t


# TODO detect type of value and call set_item_str if str or bytes?

cdef class Column_qual(Column_idx):

    @staticmethod
    def new(View    view,
            object  column_name,
            index_t nb_elements_per_line=1,
            object  elements_names=None,
            bint    dict_column=False,
            object  associated_column_name=b"",
            int     associated_column_version=-1,
            object  comments={}):
        
        return Column.new_column(view, column_name, OBI_QUAL,
                                 nb_elements_per_line=nb_elements_per_line, 
                                 elements_names=elements_names, 
                                 dict_column=dict_column, 
                                 tuples=False,
                                 associated_column_name=associated_column_name,
                                 associated_column_version=associated_column_name,
                                 comments=comments)


    cpdef object get_line(self, index_t line_nb):
        global obi_errno
        cdef   const uint8_t* value
        cdef   int            value_length
        cdef   object         result 
        value = obi_get_qual_int_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0, &value_length)
        obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem getting a value from a column")
        if value == OBIQual_int_NA :
            result = None
        else :
            result = []
            for i in range(value_length) :
                result.append(<int>value[i])            
        return result

    
    cpdef object get_bytes_line(self, index_t line_nb, int offset=-1):
        global obi_errno
        cdef   char*  value
        cdef   object result
        cdef   int    i
        value = obi_get_qual_char_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0, offset)
        obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem getting a value from a column")
        if value == OBIQual_char_NA :
            result = None
        else :
            result = value
            free(value)
        return result
    
    
    cpdef set_line(self, index_t line_nb, object value):
        global obi_errno
        cdef   uint8_t* value_b
        cdef   int      value_length
        if value is None :
            if obi_set_qual_int_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0, OBIQual_int_NA, 0) < 0:
                obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem setting a value in a column")
        else :
            value_length = len(value)
            value_b = <uint8_t*> malloc(value_length * sizeof(uint8_t))
            for i in range(value_length) :
                value_b[i] = <uint8_t>value[i]
            if obi_set_qual_int_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0, value_b, value_length) < 0:
                obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem setting a value in a column")
            free(value_b)
 
 
    cpdef set_bytes_line(self, index_t line_nb, object value, int offset=-1):
        global obi_errno
        cdef   bytes value_b        
        if value is None :
            if obi_set_qual_char_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0, OBIQual_char_NA, offset) < 0:
                obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem setting a value in a column")
        else :
            value_b = tobytes(value)
            if obi_set_qual_char_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0, value_b, offset) < 0:
                obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem setting a value in a column")

 
cdef class Column_multi_elts_qual(Column_multi_elts_idx):
 
    cpdef object get_item(self, index_t line_nb, object elt_id):
        global obi_errno
        cdef   const uint8_t*  value
        cdef   int             value_length
        cdef   object          result
        cdef   int             i
        if type(elt_id) == int :
            value = obi_get_qual_int_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id, &value_length)
        else :
            elt_name = tobytes(elt_id)
            value = obi_get_qual_int_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name, &value_length)
        obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem getting a value from a column")
        if value == OBIQual_int_NA :
            result = None
        else :
            result = []
            for i in range(value_length) :
                result.append(<int>value[i])            
        return result

    
    cpdef object get_bytes_item(self, index_t line_nb, object elt_id, int offset=-1):
        global obi_errno
        cdef   char*  value
        cdef   object result
        if type(elt_id) == int :
            value = obi_get_qual_char_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id, offset)
        else :
            elt_name = tobytes(elt_id)
            value = obi_get_qual_char_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name, offset)
        obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem getting a value from a column")
        if value == OBIQual_char_NA :
            result = None
        else :
            result = value
            free(value)
        return result

    
    cpdef object get_line(self, index_t line_nb) :
        global obi_errno
        cdef   const uint8_t*  value
        cdef   int             value_length
        cdef   list            value_in_result
        cdef   dict            result
        cdef   index_t         i
        cdef   int             j
        cdef   bint            all_NA
        cdef   list            elements_names
        cdef   Obiview_p       view_p
        cdef   OBIDMS_column_p column_p
        result = {}
        all_NA = True
        view_p = self._view.pointer()
        column_p = self.pointer()
        elements_names = self._elements_names
        for i in range(self.nb_elements_per_line) :
            value = obi_get_qual_int_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, i, &value_length)
            obi_errno_to_exception(line_nb=line_nb, elt_id=i, error_message="Problem getting a value from a column")
            if value != OBIQual_int_NA :
                value_in_result = []
                for j in range(value_length) :
                    value_in_result.append(<int>value[j])  
                result[elements_names[i]] = value_in_result
                if all_NA :
                    all_NA = False
        if all_NA :
            result = None
        return result


    cpdef object get_bytes_line(self, index_t line_nb, int offset=-1) :
        global obi_errno
        cdef   char*           value
        cdef   object          value_in_result
        cdef   dict            result
        cdef   index_t         i
        cdef   bint            all_NA
        cdef   list            elements_names
        cdef   Obiview_p       view_p
        cdef   OBIDMS_column_p column_p
        result = {}
        all_NA = True
        view_p = self._view.pointer()
        column_p = self.pointer()
        elements_names = self._elements_names
        for i in range(self.nb_elements_per_line) :
            value = obi_get_qual_char_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, i, offset)
            obi_errno_to_exception(line_nb=line_nb, elt_id=i, error_message="Problem getting a value from a column")
            if value != OBIQual_char_NA :
                value_in_result = value
                free(value)
                result[elements_names[i]] = value_in_result
                if all_NA :
                    all_NA = False
        if all_NA :
            result = None
        return result
    
    
    cpdef set_item(self, index_t line_nb, object elt_id, object value):
        global obi_errno
        cdef   uint8_t* value_b
        cdef   int      value_length
        cdef   bytes    elt_name
        
        if value is None :
            value_b = OBIQual_int_NA
            value_length = 0
        else :
            value_length = len(value)
            value_b = <uint8_t*> malloc(value_length * sizeof(uint8_t))
            for i in range(value_length) :
                value_b[i] = <uint8_t>value[i]           
   
        if type(elt_id) == int :
            if obi_set_qual_int_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id, value_b, value_length) < 0 :
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")
        else :
            elt_name = tobytes(elt_id)
            if obi_set_qual_int_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name, value_b, value_length) < 0:
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")
        
        if value is not None :
            free(value_b)
    

    cpdef set_bytes_item(self, index_t line_nb, object elt_id, object value, int offset=-1):
        global obi_errno
        cdef   bytes value_b
        cdef   bytes elt_name
        
        if value is None :
            value_b = OBIQual_char_NA
        else :
            value_b = tobytes(value)
   
        if type(elt_id) == int :
            if obi_set_qual_char_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id, value_b, offset) < 0 :
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")
        else :
            elt_name = tobytes(elt_id)
            if obi_set_qual_char_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name, value_b, offset) < 0:
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")


def register_class():
    register_column_class(OBI_QUAL, False, False, Column_qual, int)    # TODO bytes?
    register_column_class(OBI_QUAL, True, False, Column_multi_elts_qual, int)    # bytes?




#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # @UnresolvedImport
from obitools3.uri.decode import open_uri
from obitools3.dms import DMS
from obitools3.utils cimport tobytes
from obitools3.dms.capi.obiview cimport QUALITY_COLUMN
from obitools3.apps.optiongroups import addMinimalInputOption

import sys
import io
from subprocess import Popen, PIPE

from cpython.exc cimport PyErr_CheckSignals

__title__="Less equivalent"


def addOptions(parser):
    
    addMinimalInputOption(parser)


def run(config):
        
    DMS.obi_atexit()
        
    # Open the input
    input = open_uri(config['obi']['inputURI'])
    if input is None:
        raise Exception("Could not read input")
    iview = input[1]
    
    process = Popen(["less"], stdin=PIPE)
    
    for seq in iview :
        PyErr_CheckSignals()
        try:
            process.stdin.write(tobytes(repr(seq)))
            process.stdin.write(b"\n")
        except (StopIteration, BrokenPipeError, IOError):
            break

    sys.stderr.close()
    process.stdin.close()
    process.wait()

    iview.close()
    input[0].close(force=True)




#cython: language_level=3

from obitools3.apps.optiongroups import addMinimalInputOption
from obitools3.uri.decode import open_uri
from obitools3.dms import DMS
from obitools3.dms.capi.obidms cimport obi_clean_dms
from obitools3.apps.config import logger
from obitools3.utils cimport tobytes

 
__title__="Clean a DMS from unfinished views and columns"
 

def addOptions(parser):
    addMinimalInputOption(parser)
  

def run(config):
    
    DMS.obi_atexit()
    
    logger("info", "obi clean_dms")
    
    dms_path = tobytes(config['obi']['inputURI'])
    if b'.obidms' in dms_path:
        dms_path = dms_path.split(b'.obidms')[0]
    if obi_clean_dms(dms_path) < 0 :
        raise Exception("Error cleaning DMS", config['obi']['inputURI'])

    logger("info", "Done.")



#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # @UnresolvedImport
from obitools3.dms import DMS
from obitools3.uri.decode import open_uri
from obitools3.apps.optiongroups import addMinimalInputOption, addTaxonomyOption
from obitools3.dms.view import RollbackException
from obitools3.apps.config import logger
from obitools3.dms.capi.obiview cimport COUNT_COLUMN
from obitools3.utils cimport tostr

from functools import reduce
import math
import time
import sys
from cpython.exc cimport PyErr_CheckSignals


__title__="Compute basic statistics for attribute values"

'''
`obi stats` computes basic statistics for attribute values of sequence records.
The sequence records can be categorized or not using one or several ``-c`` options.
By default, only the number of sequence records and the total count are computed for each category. 
Additional statistics can be computed for attribute values in each category, such as:

    - minimum value (``-m`` option) 
    - maximum value (``-M`` option) 
    - mean value (``-a`` option) 
    - variance (``-v`` option) 
    - standard deviation (``-s`` option)
    
The result is a contingency table with the different categories in rows, and the 
computed statistics in columns. 
'''


# TODO: when is the taxonomy possibly used?

 
def addOptions(parser):
    
    addMinimalInputOption(parser)
    addTaxonomyOption(parser)

    group=parser.add_argument_group('obi stats specific options')

    group.add_argument('-c','--category-attribute',
                       action="append", dest="stats:categories",
                             metavar="<Attribute Name>",
                             default=[],
                             help="Attribute used to categorize the records.")
    
    group.add_argument('-m','--min',
                       action="append", dest="stats:minimum",
                       metavar="<Attribute Name>",
                       default=[],
                       help="Compute the minimum value of attribute for each category.")
    
    group.add_argument('-M','--max',
                       action="append", dest="stats:maximum",
                       metavar="<Attribute Name>",
                       default=[],
                       help="Compute the maximum value of attribute for each category.")

    group.add_argument('-a','--mean',
                       action="append", dest="stats:mean",
                       metavar="<Attribute Name>",
                       default=[],
                       help="Compute the mean value of attribute for each category.")

    group.add_argument('-v','--variance',
                       action="append", dest="stats:var",
                       metavar="<Attribute Name>",
                       default=[],
                       help="Compute the variance of attribute for each category.")

    group.add_argument('-s','--std-dev',
                       action="append", dest="stats:sd",
                       metavar="<Attribute Name>",
                       default=[],
                       help="Compute the standard deviation of attribute for each category.")


def statistics(values, attributes, func):
    
    stat={}
    lstat={}
    
    for var in attributes:
        if var in values:
            stat[var]={}
            lstat[var]=0
            for c in values[var]:
                v = values[var][c]
                m = func(v)
                stat[var][c]=m
                lm=len(str(m))
                if lm > lstat[var]:
                    lstat[var]=lm
                
    return stat, lstat


def minimum(values, options):
    return statistics(values, options['minimum'], min)
    

def maximum(values, options):
    return statistics(values, options['maximum'], max)


def mean(values, options):
    def average(v):
        s = reduce(lambda x,y:x+y,v,0)
        return float(s)/len(v)
    return statistics(values, options['mean'], average)


def variance(v):
    if len(v)==1: 
        return 0
    s = reduce(lambda x,y:(x[0]+y,x[1]+y**2),v,(0.,0.))
    var = round(s[1]/(len(v)-1) - s[0]**2/len(v)/(len(v)-1), 5) # round to go around shady python rounding stuff when var is actually 0
    if var == -0.0: # then fix -0 to +0 if was rounded to -0
        var = 0.0
    return var


def varpop(values, options):
    return statistics(values, options['var'], variance)
    

def sd(values, options):
    def stddev(v):
        return math.sqrt(variance(v))
    return statistics(values, options['sd'], stddev)

 
def run(config):
     
    DMS.obi_atexit()
    
    logger("info", "obi stats")

    # Open the input
    input = open_uri(config['obi']['inputURI'])
    if input is None:
        raise Exception("Could not read input view")
    i_view = input[1]

    if 'taxoURI' in config['obi'] and config['obi']['taxoURI'] is not None:
        taxo_uri = open_uri(config['obi']['taxoURI'])
        if taxo_uri is None or taxo_uri[2] == bytes:
            raise Exception("Couldn't open taxonomy")
        taxo = taxo_uri[1]
    else :
        taxo = None

    statistics = set(config['stats']['minimum']) | set(config['stats']['maximum']) | set(config['stats']['mean']) | set(config['stats']['var']) | set(config['stats']['sd'])
    total = 0
    catcount={}
    totcount={}
    values={}
    lcat=0
    
    # Initialize the progress bar
    pb = ProgressBar(len(i_view), config)
    
    for i in range(len(i_view)):
        PyErr_CheckSignals()
        pb(i)
        line = i_view[i]
        
        category = []
        for c in config['stats']['categories']:
            try:
                if taxo is not None:
                    loc_env = {'sequence': line, 'line': line, 'taxonomy': taxo}
                else:
                    loc_env = {'sequence': line, 'line': line}

                v = eval(c, loc_env, line)
                
                lv=len(str(v))
                if lv > lcat:
                    lcat=lv
                category.append(v)
            except:
                category.append(None)
                if 4 > lcat:
                    lcat=4

        category=tuple(category)
        catcount[category]=catcount.get(category,0)+1
        try: 
            totcount[category]=totcount.get(category,0)+line[COUNT_COLUMN]
        except KeyError:
            totcount[category]=totcount.get(category,0)+1
        for var in statistics:
            if var in line and line[var] is not None:
                v = line[var]
                if var not in values:
                    values[var]={}
                if category not in values[var]:
                    values[var][category]=[]
                values[var][category].append(v)    

    pb(i, force=True)
    print("", file=sys.stderr)
    
    mini, lmini = minimum(values, config['stats'])
    maxi, lmaxi = maximum(values, config['stats'])
    avg, lavg = mean(values, config['stats'])
    varp, lvarp = varpop(values, config['stats'])
    sigma, lsigma = sd(values, config['stats'])

    pcat = "%%-%ds" % lcat
    if config['stats']['minimum']:
        minvar= "min_%%-%ds" % max(len(x) for x in config['stats']['minimum'])
    else:
        minvar= "%s"
        
    if config['stats']['maximum']:
        maxvar= "max_%%-%ds" % max(len(x) for x in config['stats']['maximum'])
    else:
        maxvar= "%s"
        
    if config['stats']['mean']:
        meanvar= "mean_%%-%ds" % max(len(x) for x in config['stats']['mean'])
    else:
        meanvar= "%s"
        
    if config['stats']['var']:
        varvar= "var_%%-%ds" % max(len(x) for x in config['stats']['var'])
    else:
        varvar= "%s"
        
    if config['stats']['sd']:
        sdvar= "sd_%%-%ds" % max(len(x) for x in config['stats']['sd'])
    else:
        sdvar= "%s"
        
    hcat = ""
    
    for x in config['stats']['categories']:
        hcat += pcat % x
        hcat += "\t"
    
    for x in config['stats']['minimum']:
        hcat += minvar % x
        hcat += "\t"

    for x in config['stats']['maximum']:
        hcat += maxvar % x
        hcat += "\t"

    for x in config['stats']['mean']:
        hcat += meanvar % x
        hcat += "\t"
    
    for x in config['stats']['var']:
        hcat += varvar % x
        hcat += "\t"

    for x in config['stats']['sd']:
        hcat += sdvar % x
        hcat += "\t"

    hcat += "count\ttotal" 
    
    print(hcat)
    sorted_stats = sorted(catcount.items(), key = lambda kv:(totcount[kv[0]]), reverse=True)
    for i in range(len(sorted_stats)):
        c = sorted_stats[i][0]
        for v in c:
            if type(v) == bytes:
                print(pcat % tostr(v)+"\t", end="")
            else:
                print(pcat % str(v)+"\t", end="")
        for m in config['stats']['minimum']:
            print((("%%%dd" % lmini[m]) % mini[m][c])+"\t", end="")
        for m in config['stats']['maximum']:
            print((("%%%dd" % lmaxi[m]) % maxi[m][c])+"\t", end="")
        for m in config['stats']['mean']:
            print((("%%%df" % lavg[m]) % avg[m][c])+"\t", end="")
        for m in config['stats']['var']:
            print((("%%%df" % lvarp[m]) % varp[m][c])+"\t", end="")
        for m in config['stats']['sd']:
            print((("%%%df" % lsigma[m]) % sigma[m][c])+"\t", end="")
        print("%d" %catcount[c]+"\t", end="")
        print("%d" %totcount[c]+"\t")

    input[0].close(force=True)
    
    logger("info", "Done.")



#cython: language_level=3


from libc.stdint cimport int32_t, int64_t, uint8_t

from posix.types cimport time_t


cdef extern from *:
    ctypedef char* const_char_p "const char*"


cdef extern from "encode.h" nogil:
    bint only_ATGC(const_char_p seq)
    bint only_IUPAC_DNA(const_char_p seq)
    bint is_a_DNA_seq(const_char_p seq)


cdef extern from "obitypes.h" nogil:

    enum OBIType:
        OBI_VOID,
        OBI_INT,
        OBI_FLOAT,
        OBI_BOOL,
        OBI_CHAR,
        OBI_QUAL,
        OBI_STR,
        OBI_SEQ,
        OBI_IDX

    
    ctypedef OBIType OBIType_t
    
    enum OBIBool:
        pass
    
    ctypedef OBIBool obibool_t
    ctypedef int32_t obiint_t
    ctypedef double  obifloat_t
    ctypedef char    obichar_t  
    ctypedef int64_t index_t

    ctypedef int32_t obiversion_t

    extern obiint_t      OBIInt_NA
    extern index_t       OBIIdx_NA
    extern obifloat_t    OBIFloat_NA
    extern obichar_t     OBIChar_NA
    extern obibool_t     OBIBool_NA
    extern const_char_p  OBISeq_NA
    extern const_char_p  OBIStr_NA
    extern const_char_p  OBIQual_char_NA
    extern uint8_t*      OBIQual_int_NA
    extern void*         OBITuple_NA
    
    extern obiint_t      OBI_INT_MAX

    const_char_p name_data_type(int data_type)

ctypedef OBIType_t obitype_t



#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # @UnresolvedImport
from obitools3.dms.dms cimport DMS
from obitools3.dms.capi.obidms cimport OBIDMS_p
from obitools3.dms.view import RollbackException
from obitools3.dms.capi.obiecopcr cimport obi_ecopcr
from obitools3.apps.optiongroups import addMinimalInputOption, addMinimalOutputOption, addTaxonomyOption, addNoProgressBarOption
from obitools3.uri.decode import open_uri
from obitools3.apps.config import logger
from obitools3.utils cimport tobytes, str2bytes
from obitools3.dms.view.typed_view.view_NUC_SEQS cimport View_NUC_SEQS
from obitools3.dms.view import View

from libc.stdlib  cimport malloc, free
from libc.stdint  cimport int32_t

import sys
from io import BufferedWriter


__title__="in silico PCR"


# TODO: add option to output unique ids
def addOptions(parser):

    addMinimalInputOption(parser)
    addTaxonomyOption(parser)
    addMinimalOutputOption(parser)
    addNoProgressBarOption(parser)

 
    group = parser.add_argument_group('obi ecopcr specific options')

    group.add_argument('--primer1', '-F',
                       action="store", dest="ecopcr:primer1",
                       metavar='<PRIMER>',
                       type=str,
                       required=True,
                       help="Forward primer, length must be less than or equal to 32")

    group.add_argument('--primer2', '-R',
                       action="store", dest="ecopcr:primer2",
                       metavar='<PRIMER>',
                       type=str,
                       required=True,
                       help="Reverse primer, length must be less than or equal to 32")

    group.add_argument('--error', '-e',
                       action="store", dest="ecopcr:error",
                       metavar='<ERROR>',
                       default=0,
                       type=int,
                       help="Maximum number of errors (mismatches) allowed per primer. Default: 0.")

    group.add_argument('--min-length', '-l',
                       action="store", 
                       dest="ecopcr:min-length",
                       metavar="<MINIMUM LENGTH>",
                       type=int,
                       default=0,
                       help="Minimum length of the in silico amplified DNA fragment, excluding primers.")
    
    group.add_argument('--max-length', '-L',
                       action="store", 
                       dest="ecopcr:max-length",
                       metavar="<MAXIMUM LENGTH>",
                       type=int,
                       default=0,
                       help="Maximum length of the in silico amplified DNA fragment, excluding primers.")
 
    group.add_argument('--restrict-to-taxid', '-r',
                       action="append", 
                       dest="ecopcr:restrict-to-taxid",
                       metavar="<TAXID>",
                       type=int,
                       default=[],
                       help="Only the sequence records corresponding to the taxonomic group identified "
                            "by TAXID are considered for the in silico PCR. The TAXID is an integer "
                            "that can be found in the NCBI taxonomic database.")
 
    group.add_argument('--ignore-taxid', '-i',
                       action="append", 
                       dest="ecopcr:ignore-taxid",
                       metavar="<TAXID>",
                       type=int,
                       default=[],
                       help="The sequences of the taxonomic group identified by TAXID are not considered for the in silico PCR.")

    group.add_argument('--circular', '-c',
                       action="store_true", 
                       dest="ecopcr:circular",
                       default=False,
                       help="Considers that the input sequences are circular (e.g. mitochondrial or chloroplastic DNA).")

    group.add_argument('--salt-concentration', '-a',
                       action="store", 
                       dest="ecopcr:salt-concentration",
                       metavar="<FLOAT>",
                       type=float,
                       default=0.05,
                       help="Salt concentration used for estimating the Tm. Default: 0.05.")

    group.add_argument('--salt-correction-method', '-m',
                       action="store", 
                       dest="ecopcr:salt-correction-method",
                       metavar="<1|2>",
                       type=int,
                       default=1,
                       help="Defines the method used for estimating the Tm (melting temperature) between the primers and their corresponding "
                            "target sequences. SANTALUCIA: 1, or OWCZARZY: 2. Default: 1.")

    group.add_argument('--keep-primers', '-p',
                       action="store_true", 
                       dest="ecopcr:keep-primers",
                       default=False,
                       help="Whether to keep the primers attached to the output sequences (default: the primers are cut out).")

    group.add_argument('--keep-nucs', '-D',
                       action="store", 
                       dest="ecopcr:keep-nucs",
                       metavar="<N>",
                       type=int,
                       default=0,
                       help="Keeps N nucleotides on each side of the in silico amplified sequences, "
                            "not including the primers (implying that primers are automatically kept if N > 0).")

    group.add_argument('--kingdom-mode', '-k',
                       action="store_true", 
                       dest="ecopcr:kingdom-mode",
                       default=False,
                       help="Print in the output the kingdom of the in silico amplified sequences (default: print the superkingdom).")

 

def run(config):

    cdef int32_t* restrict_to_taxids_p = NULL
    cdef int32_t* ignore_taxids_p = NULL
    
    restrict_to_taxids_len = len(config['ecopcr']['restrict-to-taxid'])
    restrict_to_taxids_p = <int32_t*> malloc((restrict_to_taxids_len + 1) * sizeof(int32_t))   # +1 for the -1 flagging the end of the array
    for i in range(restrict_to_taxids_len) :
        restrict_to_taxids_p[i] = config['ecopcr']['restrict-to-taxid'][i]
    restrict_to_taxids_p[restrict_to_taxids_len] = -1

    ignore_taxids_len = len(config['ecopcr']['ignore-taxid'])
    ignore_taxids_p = <int32_t*> malloc((ignore_taxids_len + 1) * sizeof(int32_t))   # +1 for the -1 flagging the end of the array
    for i in range(ignore_taxids_len) :
        ignore_taxids_p[i] = config['ecopcr']['ignore-taxid'][i]
    ignore_taxids_p[ignore_taxids_len] = -1
    
    DMS.obi_atexit()
    
    logger("info", "obi ecopcr")

    # Open the input: only the DMS
    input = open_uri(config['obi']['inputURI'],
                     dms_only=True)
    if input is None:
        raise Exception("Could not read input")
    i_dms = input[0]
    i_dms_name = input[0].name
    i_view_name = input[1]

    # Open the output: only the DMS
    output = open_uri(config['obi']['outputURI'],
                      input=False,
                      dms_only=True)
    if output is None:
        raise Exception("Could not create output")
    o_dms = output[0]
    output_0 = output[0]
    o_dms_name = output[0].name
    o_view_name = output[1]

    # Open the taxonomy DMS
    taxdms = open_uri(config['obi']['taxoURI'],
                     dms_only=True)
    if taxdms is None:
        raise Exception("Could not open taxonomy DMS")
    tax_dms = taxdms[0]
    tax_dms_name = taxdms[0].name
    
    # Read taxonomy name    
    taxonomy_name = config['obi']['taxoURI'].split("/")[-1]   # Robust in theory

    # If stdout output create a temporary view in the input dms that will be deleted afterwards.
    if type(output_0)==BufferedWriter:
        o_dms = i_dms
        o_view_name = b"temp"
        while o_view_name in i_dms:   # Making sure view name is unique in input DMS
            o_view_name = o_view_name+b"_"+str2bytes(str(i))
            i+=1
    
    # Save command config in View comments
    command_line = " ".join(sys.argv[1:])
    input_dms_name=[i_dms_name]
    input_view_name= [i_view_name]
    input_dms_name.append(config['obi']['taxoURI'].split("/")[-3])
    input_view_name.append("taxonomy/"+config['obi']['taxoURI'].split("/")[-1])

    comments = View.print_config(config, "ecopcr", command_line, input_dms_name=input_dms_name, input_view_name=input_view_name)

    # TODO: primers in comments?

    if obi_ecopcr(i_dms.name_with_full_path, tobytes(i_view_name), 
                  tax_dms.name_with_full_path, tobytes(taxonomy_name), \
                  o_dms.name_with_full_path, tobytes(o_view_name), comments, \
                  tobytes(config['ecopcr']['primer1']), tobytes(config['ecopcr']['primer2']), \
                  config['ecopcr']['error'], \
                  config['ecopcr']['min-length'], config['ecopcr']['max-length'], \
                  restrict_to_taxids_p, ignore_taxids_p, \
                  config['ecopcr']['circular'], config['ecopcr']['salt-concentration'], config['ecopcr']['salt-correction-method'], \
                  config['ecopcr']['keep-nucs'], config['ecopcr']['keep-primers'], config['ecopcr']['kingdom-mode']) < 0:
        raise Exception("Error running ecopcr")

    # Save command config in DMS comments
    o_dms.record_command_line(command_line)

    free(restrict_to_taxids_p)
    free(ignore_taxids_p)

    # stdout output: write to buffer
    if type(output_0)==BufferedWriter:
        logger("info", "Printing to output...")
        o_view = o_dms[o_view_name]
        o_view.print_to_output(output_0, noprogressbar=config['obi']['noprogressbar'])
        o_view.close()

    #print("\n\nOutput view:\n````````````", file=sys.stderr)
    #print(repr(o_dms[o_view_name]), file=sys.stderr)

    # If stdout output, delete the temporary result view in the input DMS
    if type(output_0)==BufferedWriter:
        View.delete_view(i_dms, o_view_name)

    i_dms.close(force=True)
    o_dms.close(force=True)

    logger("info", "Done.")



#cython: language_level=3


from ..capi.obiview cimport Obiview_p

from ..capi.obitypes cimport index_t, obitype_t

from ..object cimport OBIWrapper                          

from ..dms cimport DMS

from ..column.column cimport Column


cdef dict __OBIDMS_VIEW_CLASS__


cdef class View(OBIWrapper):
    
    cdef DMS _dms
    
    cdef inline Obiview_p pointer(self)   

    cpdef print_to_output(self, 
                          object output, 
                          bint noprogressbar=*)
        
    cpdef delete_column(self, 
                        object column_name,
                        bint delete_file=*)
    
    cpdef rename_column(self, 
                        object current_name, 
                        object new_name)
    
    cpdef Column rewrite_column_with_diff_attributes(self,
                                                     object    column_name,
                                                     obitype_t new_data_type=*,
                                                     index_t   new_nb_elements_per_line=*,
                                                     list      new_elements_names=*,
                                                     bint      rewrite_last_line=*)

    cpdef Line_selection new_selection(self,
                                       list lines=*)


cdef class View_comments(dict):
    cdef View _view
    
    
cdef class Line_selection(list):
    
    cdef View  _view
    cdef bytes _view_name
    
    cdef index_t* __build_binary_list__(self)

    cpdef View materialize(self,
                           object view_name,
                           object comments=*)


cdef class Line :

    cdef index_t _index
    cdef View    _view
    
    cpdef repr_bytes(self)


cdef register_view_class(bytes view_type_name,
                         type view_class)

cdef register_all_view_classes()



#cython: language_level=3

from obitools3.utils cimport bytes2str, str2bytes, tostr, tobytes

from obitools3.dms.view.view cimport View

from obitools3.dms.column.column cimport Column

from .capi.obitypes cimport index_t, OBI_QUAL

from .capi.obiview cimport NUC_SEQUENCE_COLUMN, \
                           ID_COLUMN, \
                           DEFINITION_COLUMN, \
                           QUALITY_COLUMN, \
                           COUNT_COLUMN

from .capi.obiutils cimport reverse_complement_sequence

from obitools3.format.fasta import FastaFormat
from obitools3.format.fastq import FastqFormat


from cpython cimport array
import array
from copy import deepcopy
import math


SEQUENCE_COLUMN = b"SEQ"

SPECIAL_COLUMNS = [NUC_SEQUENCE_COLUMN, SEQUENCE_COLUMN, ID_COLUMN, DEFINITION_COLUMN, QUALITY_COLUMN]

QUALITY_ASCII_BASE = 33


cdef class Seq(dict) :
    def __init__(self, object id, object seq, object definition=None, object tags=None) :
        cdef object k 
        cdef bytes k_b
        self.id = id
        self.seq = seq
        self.definition = definition
        self._index = -1
        if tags is not None :
            for k in tags:
                k_b = tobytes(k)
                if k_b not in SPECIAL_COLUMNS:
                    # TODO discuss convert value to bytes if str
                    if type(tags[k]) == str:
                        self[k_b] = str2bytes(tags[k])
                    else:
                        self[k_b] = tags[k]

    @staticmethod
    def new_from_stored(Seq_Stored seq_to_clone) :
        cdef Seq  new_seq
        new_seq = Seq(seq_to_clone.id, seq_to_clone.seq, definition=seq_to_clone.definition, quality=seq_to_clone.quality, tags=seq_to_clone)
        new_seq._index = seq_to_clone._index
        return new_seq

    def __contains__(self, object key):
        return dict.__contains__(self, tobytes(key))

    def __getitem__(self, object ref):
        if type(ref) == int:
            return self.get_symbol_at(ref)
        elif type(ref) == slice:
            return self.get_slice(ref)
        else:
            return super(Seq, self).__getitem__(tobytes(ref))

    def __setitem__(self, object ref, object item):
            super(Seq, self).__setitem__(tobytes(ref), item)

    def __str__(self):
        return self.get_str()

    cpdef str get_str(self):
        return tostr(self.seq)
    
    def __len__(self):
        return len(self.seq)

    cpdef object clone(self):
        cdef object seq_class
        seq_class = type(self)
        seq = seq_class(self.id, self.seq, definition=self.definition, quality=self.quality, tags=self)
        return seq

    cpdef object get_slice(self, slice slice_to_get):
        cdef object new_seq
        cdef list   sliced_quality
        cdef type   seq_class
        seq_class = type(self)
        if QUALITY_COLUMN in self:
            sliced_quality = self.quality[slice_to_get]
        else:
            sliced_quality = None
        new_seq = seq_class(self.id+b"_SUB", self.seq[slice_to_get], definition=self.definition, quality=sliced_quality, tags=self) # TODO discuss suffix
        return new_seq

    cpdef get_symbol_at(self, int pos):
        return self.seq[pos:pos+1]

    # sequence id property getter and setter
    @property
    def id(self):  # @ReservedAssignment
        return self[ID_COLUMN]
    
    @id.setter
    def id(self, object new_id):  # @ReservedAssignment @DuplicatedSignature
        self[ID_COLUMN] = tobytes(new_id)
        
    # sequence property getter and setter
    @property
    def seq(self):
        return self[SEQUENCE_COLUMN]
    
    @seq.setter
    def seq(self, object new_seq):  # @DuplicatedSignature
        self[SEQUENCE_COLUMN] = tobytes(new_seq)
        
    # sequence definition property getter and setter
    @property
    def definition(self):
        return self[DEFINITION_COLUMN]
    
    @definition.setter
    def definition(self, object new_definition):  # @DuplicatedSignature
        if new_definition is not None:
            new_definition = tobytes(new_definition)
        self[DEFINITION_COLUMN] = new_definition

    # sequence index (for reference in a view eventually) property getter and setter
    @property
    def index(self):  # @ReservedAssignment
        return self._index

    @index.setter
    def index(self, int idx):  # @DuplicatedSignature
        self._index = idx

        
cdef class Nuc_Seq(Seq) :

    def __init__(self, object id, object seq, object definition=None, object quality=None, int offset=QUALITY_ASCII_BASE, object tags=None) :
        cdef object k 
        cdef bytes  k_b
        cdef int    q
        cdef list   q_proba_list
                
        self.id = id
        self.seq = seq
        self.definition = definition
        if quality is not None:
            self.set_quality(quality, offset=offset)
        
        self._is_revcomp = False
        
        if tags is not None:
            for k in tags:
                k_b = tobytes(k)
                if k_b not in SPECIAL_COLUMNS:
                    # TODO discuss convert value to bytes if str
                    if type(tags[k]) == str:
                        self[k_b] = str2bytes(tags[k])
                    else:
                        self[k_b] = tags[k]

    @staticmethod
    def new_from_stored(Nuc_Seq_Stored seq_to_clone) :
        cdef Nuc_Seq new_seq
        new_seq = Nuc_Seq(seq_to_clone.id, seq_to_clone.seq, definition=seq_to_clone.definition, quality=seq_to_clone.quality, tags=seq_to_clone)
        new_seq._index = seq_to_clone.index
        return new_seq

    # is_revcomp property getter and setter (boolean indicating whether the sequence was created by reverse complementing another sequence)
    @property
    def is_revcomp(self):
        return self._is_revcomp

    @is_revcomp.setter
    def is_revcomp(self, bint is_revcomp):  # @DuplicatedSignature
        self._is_revcomp = is_revcomp

    # nuc sequence property getter and setter
    @property
    def seq(self):
        return self[NUC_SEQUENCE_COLUMN]
    
    @seq.setter
    def seq(self, object new_seq):  # @DuplicatedSignature
        self[NUC_SEQUENCE_COLUMN] = tobytes(new_seq)
        
    # sequence quality property getter and setter
    @property
    def quality(self):
        if QUALITY_COLUMN in self:
            return self[QUALITY_COLUMN]
        else:
            return None
    
    cpdef set_quality(self, object new_quality, int offset=QUALITY_ASCII_BASE):
        cdef list quality_int
        if type(new_quality) == list:
            quality_int = new_quality
        elif type(new_quality) == str or type(new_quality) == bytes:
            quality_int = []
            for pos in range(len(new_quality)):
                quality_int.append(ord(new_quality[pos:pos+1])-offset)
        else:
            raise Exception("Sequence quality in unrecognized format:", type(new_quality))
        self[QUALITY_COLUMN] = quality_int

    # sequence quality array property getter and setter
    @property
    def quality_array(self):
        if self._quality_array is None:
            if QUALITY_COLUMN in self:
                self._quality_array = self.build_quality_array(self[QUALITY_COLUMN])
        return self._quality_array

    cpdef object build_quality_array(self, list quality):
        cdef int    q
        cdef list   q_proba_list
        cdef object qual_array
        q_proba_list = [(10.**(-q/10.)) for q in quality]
        return array.array('d', q_proba_list)

    # reverse complement property getter
    @property
    def reverse_complement(self):
        cdef bytes rev_comp
        cdef object seq_class
        if self._reverse_complement is None:
            rev_comp = self.build_reverse_complement()
            if QUALITY_COLUMN in self:
                reversed_quality = self.quality[::-1]
            else:
                reversed_quality = None
            seq = Nuc_Seq(self.id+b"_CMP", rev_comp, definition=self.definition, quality=reversed_quality, tags=self)
            seq.is_revcomp = True
            self._reverse_complement = seq
        return self._reverse_complement
    
    cpdef bytes build_reverse_complement(self) :
        cdef bytes rev_comp
        rev_comp = deepcopy(self[NUC_SEQUENCE_COLUMN]).lower()
        return <bytes>reverse_complement_sequence(rev_comp)        


cdef class Seq_Stored(Line) :

    def __getitem__(self, object ref):
        if type(ref) == int:
            return self.get_symbol_at(ref)
        elif type(ref) == slice:
            return self.get_slice(ref)
        else:
            return super(Seq_Stored, self).__getitem__(ref)

    cpdef object get_slice(self, slice slice_to_get):
        cdef object new_seq
        cdef list   sliced_quality
        cdef type   seq_class
        seq_class = type(self)
        if QUALITY_COLUMN in self:
            sliced_quality = self.quality[slice_to_get]
        else:
            sliced_quality = None
        new_seq = seq_class(self.id+b"_SUB", self.seq[slice_to_get], definition=self.definition, quality=sliced_quality, tags=self)  # TODO discuss suffix
        return new_seq

    cpdef get_symbol_at(self, int pos):
        return self.seq[pos:pos+1]

    # sequence id property getter and setter
    @property
    def id(self):               # @ReservedAssignment @DuplicatedSignature
        return self._view.get_column(ID_COLUMN).get_line(self.index)
    
    @id.setter
    def id(self, object new_id):   # @ReservedAssignment @DuplicatedSignature
        self._view.get_column(ID_COLUMN).set_line(self._index, tobytes(new_id))
        
    # sequence definition property getter and setter
    @property
    def definition(self):
        return self._view.get_column(DEFINITION_COLUMN).get_line(self.index)
    
    @definition.setter
    def definition(self, object new_def): # @DuplicatedSignature
        self._view.get_column(DEFINITION_COLUMN).set_line(self._index, tobytes(new_def))
    

cdef class Nuc_Seq_Stored(Seq_Stored) :
    
#    def __init__(self, View view, index_t line_nb) :
#        self = Seq_Stored.__init__(self, view, line_nb)
        #self._reverse_complement = None    # TODO unnecessary?
    
#     def __setitem__(self, object column_name, object value):
#         try:
#             super(Seq, self).__setitem__(column_name, value)
#         except CantGuessTypeException:
#             column_name_b = tobytes(column_name)
#             if column_name_b in SPECIAL_COLUMNS:
#                 self[column_name_b] = tags[column_name]

    cpdef set(self, object id, object seq, object definition=None, object quality=None, int offset=QUALITY_ASCII_BASE, object tags=None):
        cdef object k 
        cdef bytes  k_b
        cdef int    q
        cdef list   q_proba_list
        
        self[ID_COLUMN] = tobytes(id)
        self[NUC_SEQUENCE_COLUMN] = tobytes(seq)
        if definition is not None:
            self.definition = tobytes(definition)
        if quality is not None:
            if type(quality) == list:
                self.set_quality_int(quality)
            elif type(quality) == str or type(quality) == bytes:
                self.set_quality_char(quality, offset=offset)
            else:
                raise Exception("Sequence quality in unrecognized format")
            
        if tags is not None:
            for k in tags:
                k_b = tobytes(k)
                if k_b not in SPECIAL_COLUMNS:
                    if type(tags[k]) == str:
                        self[k_b] = str2bytes(tags[k])
                    else:
                        self[k_b] = tags[k]

    # seq property getter and setter
    @property
    def seq(self):
        if not self._view.read_only :
            return self._view.get_column(NUC_SEQUENCE_COLUMN).get_line(self.index)
        else:
            if self._seq is None:
                self._seq = self._view.get_column(NUC_SEQUENCE_COLUMN).get_line(self.index)
            return self._seq
    
    @seq.setter
    def seq(self, object new_seq): # @DuplicatedSignature
        self._view.get_column(NUC_SEQUENCE_COLUMN).set_line(self.index, tobytes(new_seq))

    cpdef set_quality_int(self, list new_qual):
        if QUALITY_COLUMN not in self:
            Column.new_column(self._view, QUALITY_COLUMN, OBI_QUAL)
        self._view.get_column(QUALITY_COLUMN).set_line(self.index, new_qual)

    cpdef set_quality_char(self, object new_qual, int offset=QUALITY_ASCII_BASE):
        if QUALITY_COLUMN not in self:
            Column.new_column(self._view, QUALITY_COLUMN, OBI_QUAL)
        self._view.get_column(QUALITY_COLUMN).set_str_line(self.index, tobytes(new_qual), offset=offset)

    # quality property getter and setter
    @property
    def quality(self):
        if QUALITY_COLUMN in self:
            return self._view.get_column(QUALITY_COLUMN).get_line(self.index)
        else:
            return None
    
    @quality.setter
    def quality(self, object new_qual): # @DuplicatedSignature
        if (new_qual is None) or (type(new_qual) == list) :     # TODO check that quality column exists
            self.set_quality_int(new_qual)
        # WARNING: default offset used
        elif (type(new_qual) == str) or (type(new_qual) == bytes)  :  # Quality is in str form
            self.set_quality_char(new_qual)
        else :
            raise Exception("Sequence quality in unrecognized format")

    # quality character string property getter
    # WARNING: default offset used
    @property
    def quality_bytes(self):
        if QUALITY_COLUMN in self:
            return self._view.get_column(QUALITY_COLUMN).get_bytes_line(self._index)
        else:
            return None

    # sequence quality array property getter and setter
    @property
    def quality_array(self):
        if self._quality_array is None:
            if QUALITY_COLUMN in self:
                self._quality_array = self.build_quality_array(self._view.get_column(QUALITY_COLUMN).get_line(self.index))
        return self._quality_array

    cpdef object build_quality_array(self, list quality):
        cdef int    q
        cdef list   q_proba_list
        cdef object qual_array
        q_proba_list = [(10.**(-q/10.)) for q in quality]
        return array.array('d', q_proba_list)

    # reverse complement property getter
    @property
    def reverse_complement(self):
        cdef bytes rev_comp
        cdef list  reversed_quality
        if self._reverse_complement is None:
            rev_comp = self.build_reverse_complement()
            if QUALITY_COLUMN in self:
                reversed_quality = self.quality[::-1]
            else:
                reversed_quality = None
            seq = Nuc_Seq(self.id+b"_CMP", rev_comp, definition=self.definition, quality=reversed_quality, tags=self)
            seq.is_revcomp = True
            self._reverse_complement = seq
        return self._reverse_complement
    
    cpdef bytes build_reverse_complement(self) :
        cdef bytes rev_comp
        rev_comp = deepcopy(self[NUC_SEQUENCE_COLUMN]).lower()
        return <bytes>reverse_complement_sequence(rev_comp)
    
    def __str__(self):
        return self.get_str()

    cpdef str get_str(self):
        return tostr(self._view.get_column(NUC_SEQUENCE_COLUMN).get_line(self.index))

    def __len__(self):
        return len(self._view.get_column(NUC_SEQUENCE_COLUMN).get_line(self.index))

    def __repr__(self):
        return bytes2str(self.repr_bytes())
    
    cpdef repr_bytes(self):
        if self.quality is None:
            formatter = FastaFormat()
        else:
            formatter = FastqFormat()
        return formatter(self)




#cython: language_level=3

from obitools3.apps.progress cimport ProgressBar  # @UnresolvedImport
from obitools3.dms import DMS
from obitools3.dms.view.view cimport View, Line_selection
from obitools3.uri.decode import open_uri
from obitools3.apps.optiongroups import addMinimalInputOption, addTaxonomyOption, addMinimalOutputOption, addNoProgressBarOption
from obitools3.dms.view import RollbackException
from obitools3.apps.config import logger
from obitools3.utils cimport tobytes, str2bytes

from functools import reduce
import time
import re
import sys
import ast
from io import BufferedWriter
from cpython.exc cimport PyErr_CheckSignals

 
__title__="Grep view lines that match the given predicates"


# TODO should sequences that have a grepped attribute at None be grepped or not? (in obi1 they are but....)

 
def addOptions(parser):
    
    addMinimalInputOption(parser)
    addTaxonomyOption(parser)
    addMinimalOutputOption(parser)
    addNoProgressBarOption(parser)

    group=parser.add_argument_group("obi grep specific options")
 
    group.add_argument("--predicate", "-p",
                       action="append", dest="grep:grep_predicates",
                       metavar="<PREDICATE>",
                       default=[],
                       type=str,
                       help="Python boolean expression to be evaluated in the "
                            "sequence/line context. The attribute name can be "
                            "used in the expression as a variable name. "
                            "An extra variable named 'sequence' or 'line' refers "
                            "to the sequence or line object itself. "
                            "Several -p options can be used on the same "
                            "command line.")
 
    group.add_argument("-S", "--sequence",
                       action="store", dest="grep:seq_pattern",
                       metavar="<REGULAR_PATTERN>",
                       type=str,
                       help="Regular expression pattern used to select "
                            "the sequence. The pattern is case insensitive.")
    
    group.add_argument("-D", "--definition",
                       action="store", dest="grep:def_pattern",
                       metavar="<REGULAR_PATTERN>",
                       type=str,
                       help="Regular expression pattern used to select "
                            "the definition of the sequence. The pattern is case insensitive.")
    
    group.add_argument("-I", "--identifier",
                       action="store", dest="grep:id_pattern",
                       metavar="<REGULAR_PATTERN>",
                       type=str,
                       help="Regular expression pattern used to select "
                            "the identifier of the sequence. The pattern is case insensitive.")
    
    group.add_argument("--id-list",
                       action="store", dest="grep:id_list",
                       metavar="<FILE_NAME>",
                       type=str,
                       help="File containing the identifiers of the sequences to select.")

    group.add_argument("-a", "--attribute",
                       action="append", dest="grep:attribute_patterns",
                       type=str,
                       default=[],
                       metavar="<ATTRIBUTE_NAME>:<REGULAR_PATTERN>",
                       help="Regular expression pattern matched against "
                            "the attributes of the sequence. "
                            "The pattern is case sensitive. "
                            "Several -a options can be used on the same "
                            "command line.")
    
    group.add_argument("-A", "--has-attribute",
                       action="append", dest="grep:attributes",
                       type=str,
                       default=[],
                       metavar="<ATTRIBUTE_NAME>",
                       help="Select records with the attribute <ATTRIBUTE_NAME> "
                            "defined (not set to NA value). "
                            "Several -A options can be used on the same "
                            "command line.")
    
    group.add_argument("-L", "--lmax",
                       action="store", dest="grep:lmax",
                       metavar="<MAX_LENGTH>",
                       type=int,
                       help="Keep sequences shorter than MAX_LENGTH.")
                             
    group.add_argument("-l", "--lmin",
                       action="store", dest="grep:lmin",
                       metavar="<MIN_LENGTH>",
                       type=int,
                       help="Keep sequences longer than MIN_LENGTH.")
    
    group.add_argument("-v", "--invert-selection",
                       action="store_true", dest="grep:invert_selection",
                       default=False,
                       help="Invert the selection.")


    group=parser.add_argument_group("Taxonomy filtering specific options")  #TODO put somewhere else? not in grep
 
    group.add_argument('--require-rank',
                       action="append", dest="grep:required_ranks",
                       metavar="<RANK_NAME>",
                       type=str,
                       default=[],
                       help="Select sequences with a taxid that is or has "
                            "a parent of rank <RANK_NAME>.")
     
    group.add_argument('-r', '--required',
                       action="append", dest="grep:required_taxids",
                       metavar="<TAXID>",
                       type=int,
                       default=[],
                       help="Select the sequences having the ancestor of taxid <TAXID>. "
                            "If several ancestors are specified (with \n'-r taxid1 -r taxid2'), "
                            "the sequences having at least one of them are selected.")
     
    # TODO useless option equivalent to -r -v?
    group.add_argument('-i','--ignore',
                     action="append", dest="grep:ignored_taxids",
                     metavar="<TAXID>",
                     type=int,
                     default=[],
                     help="Ignore the sequences having the ancestor of taxid <TAXID>. "
                          "If several ancestors are specified (with \n'-r taxid1 -r taxid2'), "
                          "the sequences having at least one of them are ignored.")


def obi_compile_eval(str expr):
    
    class MyVisitor(ast.NodeTransformer):
        def visit_Str(self, node: ast.Str):
            result = ast.Bytes(s = node.s.encode('utf-8'))
            return ast.copy_location(result, node)
   
    expr = "obi_eval_result="+expr
    tree = ast.parse(expr)
    optimizer = MyVisitor()
    tree = optimizer.visit(tree)
    return compile(tree, filename="<ast>", mode="exec")

    
def obi_eval(compiled_expr, loc_env, line):

    exec(compiled_expr, {}, loc_env)
    obi_eval_result = loc_env["obi_eval_result"]
    return obi_eval_result
    

def Filter_generator(options, tax_filter, i_view):

    # Initialize conditions
    predicates = None
    if "grep_predicates" in options:
        predicates = [obi_compile_eval(p) for p in options["grep_predicates"]]
    attributes = None
    if "attributes" in options and len(options["attributes"]) > 0:
        attributes = options["attributes"]
        for attribute in attributes:
            if attribute not in i_view:
                return None
    lmax = None
    if "lmax" in options:
        lmax = options["lmax"]
    lmin = None
    if "lmin" in options:
        lmin = options["lmin"]
    invert_selection = options["invert_selection"]
    id_set = None
    if "id_list" in options:
        id_set = set(x.strip() for x in open(options["id_list"], 'rb'))

    # Initialize the regular expression patterns
    seq_pattern = None
    if "seq_pattern" in options:
        seq_pattern = re.compile(tobytes(options["seq_pattern"]), re.I)
    id_pattern = None
    if "id_pattern" in options:
        id_pattern = re.compile(tobytes(options["id_pattern"]))
    def_pattern = None
    if "def_pattern" in options:
        def_pattern = re.compile(tobytes(options["def_pattern"]))
    attribute_patterns={}
    if "attribute_patterns" in options and len(options["attribute_patterns"]) > 0:
        for p in options["attribute_patterns"]:
            attribute, pattern = p.split(":", 1)
            if attribute not in i_view:
                return None
            attribute_patterns[tobytes(attribute)] = re.compile(tobytes(pattern))
    
    def filter(line, loc_env):
        cdef bint good = True
        
        if seq_pattern and hasattr(line, "seq"):
            good = <bint>(seq_pattern.search(line.seq))
        
        if good and id_pattern and hasattr(line, "id"):
            good = <bint>(id_pattern.search(line.id))
            
        if good and id_set is not None and hasattr(line, "id"):
            good = line.id in id_set
            
        if good and def_pattern and hasattr(line, "definition"):
            good = <bint>(def_pattern.search(line.definition))
             
        if good and attributes:  # TODO discuss that we test not None
            good = reduce(lambda bint x, bint y: x and y,
                           (line[attribute] is not None for attribute in attributes),
                           True)
             
        if good and attribute_patterns:
            good = (reduce(lambda bint x, bint y : x and y, 
                        (line[attribute] is not None for attribute in attribute_patterns),
                        True)
                    and
                    reduce(lambda bint x, bint y: x and y,
                        (<bint>(attribute_patterns[attribute].search(tobytes(str(line[attribute]))))
                        for attribute in attribute_patterns), 
                        True)
                   )
                     
        if good and predicates:
            good = (reduce(lambda bint x, bint y: x and y,
                    (bool(obi_eval(p, loc_env, line))
                    for p in predicates), True))

        if good and lmin:
            good = len(line) >= lmin
             
        if good and lmax:
            good = len(line) <= lmax
         
        if good:
            good = tax_filter(line)
       
        if invert_selection :
            good = not good

        return good
    
    return filter


def Taxonomy_filter_generator(taxo, options):
    
    if (("required_ranks" in options and options["required_ranks"]) or \
       ("required_taxids" in options and options["required_taxids"]) or \
       ("ignored_taxids" in options and options["ignored_taxids"])) and \
       (taxo is None):
        raise RollbackException("obi grep error: can't use taxonomy options without providing a taxonomy. Rollbacking view")
    
    if taxo is not None:
        def tax_filter(seq):
            good = True
            if b'TAXID' in seq and seq[b'TAXID'] is not None:   # TODO use macro
                taxid = seq[b'TAXID']
                if "required_ranks" in options and options["required_ranks"]:
                    taxon_at_rank = reduce(lambda x,y: x and y,
                                           (taxo.get_taxon_at_rank(seq[b'TAXID'], rank) is not None
                                            for rank in options["required_ranks"]),
                                           True)
                    good = good and taxon_at_rank 
                if "required_taxids" in options and options["required_taxids"]:
                    good = good and reduce(lambda x,y: x or y,
                                           (taxo.is_ancestor(r, taxid) 
                                            for r in options["required_taxids"]),
                                           False)
                if "ignored_taxids" in options and options["ignored_taxids"]:
                    good = good and not reduce(lambda x,y: x or y,
                                               (taxo.is_ancestor(r,taxid) 
                                                for r in options["ignored_taxids"]),
                                               False)
            return good
    else:
        def tax_filter(seq):
            return True
    return tax_filter


def run(config):
     
    DMS.obi_atexit()
    
    logger("info", "obi grep")

    # Open the input
    input = open_uri(config["obi"]["inputURI"])
    if input is None:
        raise Exception("Could not read input view")
    i_dms = input[0]
    i_view = input[1]

    # Open the output: only the DMS
    output = open_uri(config['obi']['outputURI'],
                      input=False,
                      dms_only=True)
    if output is None:
        raise Exception("Could not create output view")
    o_dms = output[0]
    output_0 = output[0]
    final_o_view_name = output[1]

    # If stdout output or the input and output DMS are not the same, create a temporary view that will be exported and deleted afterwards.
    if i_dms != o_dms or type(output_0)==BufferedWriter:
        temporary_view_name = b"temp"
        i=0
        while temporary_view_name in i_dms:  # Making sure view name is unique in input DMS
            temporary_view_name = temporary_view_name+b"_"+str2bytes(str(i))
            i+=1
        o_view_name = temporary_view_name
        if type(output_0)==BufferedWriter:
            o_dms = i_dms
    else:
        o_view_name = final_o_view_name

    if 'taxoURI' in config['obi'] and config['obi']['taxoURI'] is not None:
        taxo_uri = open_uri(config["obi"]["taxoURI"])
        if taxo_uri is None or taxo_uri[2] == bytes:
            raise Exception("Couldn't open taxonomy")
        taxo = taxo_uri[1]
    else :
        taxo = None

    # Initialize the progress bar
    if config['obi']['noprogressbar'] == False:
        pb = ProgressBar(len(i_view), config)
    else:
        pb = None
 
    # Apply filter
    tax_filter = Taxonomy_filter_generator(taxo, config["grep"])
    filter = Filter_generator(config["grep"], tax_filter, i_view)        
    selection = Line_selection(i_view)
    
    if filter is None and config["grep"]["invert_selection"]: # all sequences are selected: filter is None if no line will be selected because some columns don't exist
        for i in range(len(i_view)):
            PyErr_CheckSignals()
            if pb is not None:
                pb(i)
            selection.append(i)
    
    elif filter is not None :   # filter is None if no line will be selected because some columns don't exist
        for i in range(len(i_view)):
            PyErr_CheckSignals()
            if pb is not None:
                pb(i)
            line = i_view[i]
            
            loc_env = {"sequence": line, "line": line, "taxonomy": taxo, "obi_eval_result": False}
            
            good = filter(line, loc_env)
     
            if good :
                selection.append(i)
   
    if pb is not None:
        pb(len(i_view), force=True)
        print("", file=sys.stderr)

    # Create output view with the line selection
    try:
        o_view = selection.materialize(o_view_name)
    except Exception, e:
        raise RollbackException("obi grep error, rollbacking view: "+str(e), o_view)

    logger("info", "Grepped %d entries" % len(o_view))

    # Save command config in View and DMS comments
    command_line = " ".join(sys.argv[1:])
    input_dms_name=[input[0].name]
    input_view_name=[input[1].name]
    if 'taxoURI' in config['obi'] and config['obi']['taxoURI'] is not None:
        input_dms_name.append(config['obi']['taxoURI'].split("/")[-3])
        input_view_name.append("taxonomy/"+config['obi']['taxoURI'].split("/")[-1])
    o_view.write_config(config, "grep", command_line, input_dms_name=input_dms_name, input_view_name=input_view_name)
    o_dms.record_command_line(command_line)

    # If input and output DMS are not the same, export the temporary view to the output DMS
    # and delete the temporary view in the input DMS
    if i_dms != o_dms:
        o_view.close()
        View.import_view(i_dms.full_path[:-7], o_dms.full_path[:-7], o_view_name, final_o_view_name)
        o_view = o_dms[final_o_view_name]
    
    # stdout output: write to buffer
    if type(output_0)==BufferedWriter:
        logger("info", "Printing to output...")
        o_view.print_to_output(output_0, noprogressbar=config['obi']['noprogressbar'])
        o_view.close()

    #print("\n\nOutput view:\n````````````", file=sys.stderr)
    #print(repr(o_view), file=sys.stderr)

    # If the input and the output DMS are different or if stdout output, delete the temporary imported view used to create the final view
    if i_dms != o_dms or type(output_0)==BufferedWriter:
        View.delete_view(i_dms, o_view_name)
        o_dms.close(force=True)
    i_dms.close(force=True)

    logger("info", "Done.")



#cython: language_level=3

'''
Created on June 12th 2018

@author: coissac/mercier
'''


import logging
import re
from itertools import chain
from obitools3.utils cimport str2bytes

        
_featureMatcher = re.compile(b'^(FT|  )   [^ ].+\n((FT|  )    .+\n)+',re.M)
_featureCleaner = re.compile(b'^FT',re.M)

def textFeatureIterator(fttable):
    '''
    Iterate through a textual description of a feature table in a genbank
    or embl format. Return at each step a text representation of each individual
    feature composing the table.
    
    @param fttable:  a string corresponding to the feature table of a genbank
                     or an embl entry
                      
    @type fttable: C{str}
    
    @return: an iterator on str
    @rtype: iterator
    
    @see: L{ftParser}
    '''
    for m in _featureMatcher.finditer(fttable):
        t = m.group()
        t = _featureCleaner.sub(b'  ',t)
        yield t
   
        
_qualifierMatcher = re.compile(b'(?<=^ {21}/).+(\n {21}[^/].+)*',re.M)
_qualifierCleanner= re.compile(b"^ +",re.M)
        
def qualifierIterator(qualifiers):
    '''
    Parse a textual description of a feature in embl or genbank format
    as returned by the textFeatureIterator iterator and iterate through 
    the key, value qualified defining this location.
     
    @param qualifiers: substring containing qualifiers
    @type qualifiers: str
    
    @return: an iterator on tuple (key,value), where keys are C{str}
    @rtype: iterator
    '''
    for m in _qualifierMatcher.finditer(qualifiers):
        t = m.group()
        t = _qualifierCleanner.sub(b'',t)
        t = t.split(b'=',1)
        if len(t)==1:
            t = (t[0],None)
        else:
            if t[0]==b'translation':
                value = t[1].replace(b'\n',b'')
            else:
                value = t[1].replace(b'\n',b' ')
            try:
                value = eval(value)
                if type(value) == str:
                    value = str2bytes(value)
            except:
                pass
            t = (t[0],value)
        yield t
    
     
_ftmatcher = re.compile(b'(?<=^ {5})\S+')
_qualifiersMatcher = re.compile(b'^ +/.+',re.M+re.DOTALL)

def ftParser(feature):
    fttype = _ftmatcher.search(feature).group()
    qualifiers=_qualifiersMatcher.search(feature)
    if qualifiers is not None:
        qualifiers=qualifiers.group()
    else:
        qualifiers=b""
        logging.debug("Qualifiers regex not matching on \n=====\n%s\n========" % feature)

    return fttype,qualifiers       


class Feature(dict):
    
    def __init__(self,type):
        self._fttype=type

    def getFttype(self):
        return self._fttype

    def __str__(self):
        return repr(self)
    
    def __repr__(self):
        return str((self.ftType, dict.__repr__(self)))
    
    ftType = property(getFttype, None, None, "Feature type name")
       

def featureFactory(featureDescription):
    fttype,qualifiers = ftParser(featureDescription)
    feature = Feature(fttype)
    feature.raw  = featureDescription

    for k,v in qualifierIterator(qualifiers):
        feature.setdefault(k,[]).append(v)
        
    return feature
       
        
def featureIterator(featureTable,skipError=False):
    for tft in textFeatureIterator(featureTable):
        try:
            feature = featureFactory(tft)
        except AssertionError,e:
            logging.debug("Parsing error on feature :\n===============\n%s\n===============" % tft)
            if not skipError:
                raise e
            logging.debug("\t===> Error skipped")
            continue
            
        yield feature

        
def extractTaxon(bytes text, dict tags):
         
    s = next(featureIterator(text))
    s = [s]          
    
    t = set(int(v[6:]) for v in chain(*tuple(f[b'db_xref'] for f in s if b'db_xref' in f)) 
            if  v[0:6]==b'taxon:')
    if len(t)==1 :
        taxid=t.pop()
        if taxid >=0:
            tags[b'TAXID']=taxid      

    t = set(chain(*tuple(f[b'organism'] for f in s if b'organism' in f))) 
    if len(t)==1:
        tags[b'organism']=t.pop()



#cython: language_level=3

'''
Created on 28 juillet 2017

@author: coissac
'''

from os import environb,getpid
from os.path import join, isdir
from tempfile import TemporaryDirectory, _get_candidate_names
from shutil import rmtree
from atexit import register

from obitools3.dms.dms import DMS

from obitools3.apps.config import getConfiguration
from obitools3.apps.config import logger

cpdef get_temp_dir():
    """
    Returns a temporary directory object specific of this instance of obitools.
    
    This is an application function. It cannot be called out of an obi command.
    It requires a valid configuration.
    
    If the function is called several time from the same obi session, the same
    directory is returned.
    
    If the OBITMP environment variable exist, the temporary directory is created
    inside this directory.
    
    The directory is automatically destroyed at the end of the end of the process.
    
        @return: a temporary python directory object.
    """
    cdef bytes tmpdirname
    cdef dict config = getConfiguration()

    root = config["__root_config__"]
    
    try:
        return config[root]["tempdir"].name
    except KeyError:
        pass
    
    try:
        basedir=environb[b'OBITMP']
    except KeyError:
        basedir=None
    
    tmp = TemporaryDirectory(dir=basedir)
        
    config[root]["tempdir"]=tmp
    
    return tmp.name
 
cpdef get_temp_dir_name():
    """
    Returns the name of the  temporary directory object 
    specific of this instance of obitools.
    
        @return: the name of the temporary directory.
        
        @see get_temp_dir
    """
    return get_temp_dir_name().name
          
               
cpdef get_temp_dms():

    cdef bytes tmpdirname                   # @DuplicatedSignature
    cdef dict config = getConfiguration()   # @DuplicatedSignature
    cdef DMS tmpdms
    
    root = config["__root_config__"]
    
    try:
        return config[root]["tempdms"]
    except KeyError:
        pass

    tmpdirname=get_temp_dir()
        
    tempname = join(tmpdirname,
                    b"obi.%d.%s" % (getpid(),
                                    tobytes(next(_get_candidate_names())))
                   )
        
    tmpdms = DMS.new(tempname)
    
    config[root]["tempdms"]=tmpdms
    
    return tmpdms
    
    



#cython: language_level=3

from obitools3.dms.capi.obierrno cimport obi_errno

from ..column cimport register_column_class

from ...view.view cimport View

from ..column cimport Column

from obitools3.utils cimport tobytes, \
                             obi_errno_to_exception

from ...capi.obiview cimport obi_get_str_with_elt_name_and_col_p_in_view, \
                             obi_get_str_with_elt_idx_and_col_p_in_view, \
                             obi_set_str_with_elt_name_and_col_p_in_view, \
                             obi_set_str_with_elt_idx_and_col_p_in_view, \
                             obi_get_array_with_col_p_in_view, \
                             obi_set_array_with_col_p_in_view, \
                             Obiview_p

from ...capi.obidmscolumn cimport OBIDMS_column_p

from ...capi.obitypes     cimport OBI_STR, \
                                  OBIStr_NA, \
                                  OBITuple_NA, \
                                  const_char_p

from libc.stdint cimport int32_t
from libc.stdlib  cimport calloc, free
from libc.string  cimport strcpy


cdef class Column_str(Column_idx):

    @staticmethod
    def new(View    view,
            object  column_name,
            index_t nb_elements_per_line=1,
            object  elements_names=None,
            bint    dict_column=False,
            bint    tuples=False,
            object  comments={}):
        
        return Column.new_column(view, column_name, OBI_STR,
                                 nb_elements_per_line=nb_elements_per_line, 
                                 elements_names=elements_names, 
                                 dict_column=dict_column, 
                                 tuples=tuples,
                                 comments=comments)


    cpdef object get_line(self, index_t line_nb):
        global obi_errno
        cdef   const_char_p value
        cdef   object       result
        value = obi_get_str_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0)
        obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem getting a value from a column")
        if value == OBIStr_NA :
            result = None
        # For columns containing character strings that should be evaluated:
        elif self.to_eval:
            result = eval(value)
        else :
            result = <bytes> value      # NOTE: value is not freed because the pointer points to a mmapped region in an AVL data file.
        return result


    cpdef set_line(self, index_t line_nb, object value):
        global obi_errno
        cdef   char* value_b
        cdef   bytes value_bytes
        
        if value is None :
            value_b = <char*>OBIStr_NA
        else :
            if self.name == b'sample' or self.name == b'SAMPLE':
                if type(value) == int:
                    value = str(value) # force sample ids to be str
            value_bytes = tobytes(value)
            value_b = <char*>value_bytes

        if obi_set_str_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, 0, value_b) < 0:
            obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem setting a value in a column")
 
 
cdef class Column_multi_elts_str(Column_multi_elts_idx):
 
    cpdef object get_item(self, index_t line_nb, object elt_id) :
        global obi_errno
        cdef   const_char_p value
        cdef   object       result
        if type(elt_id) == int :
            value = obi_get_str_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id)
        else :
            elt_name = tobytes(elt_id)
            value = obi_get_str_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name)
        obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem getting a value from a column")
        if value == OBIStr_NA :
            result = None
        else :
            result = <bytes> value      # NOTE: value is not freed because the pointer points to a mmapped region in an AVL data file.
        return result


    cpdef object get_line(self, index_t line_nb) :
        global obi_errno
        cdef   const_char_p    value
        cdef   object          value_in_result
        cdef   dict            result
        cdef   index_t         i
        cdef   bint            all_NA
        cdef   list            elements_names
        cdef   Obiview_p       view_p
        cdef   OBIDMS_column_p column_p
        result = {}
        all_NA = True
        view_p = self._view.pointer()
        column_p = self.pointer()
        elements_names = self._elements_names
        for i in range(self.nb_elements_per_line) :
            value = obi_get_str_with_elt_idx_and_col_p_in_view(view_p, column_p, line_nb, i)
            obi_errno_to_exception(line_nb=line_nb, elt_id=i, error_message="Problem getting a value from a column")
            if value != OBIStr_NA :
                value_in_result = <bytes> value
                result[elements_names[i]] = value_in_result
                if all_NA :
                    all_NA = False
        if all_NA :
            result = None
        return result


    cpdef set_item(self, index_t line_nb, object elt_id, object value):
        global obi_errno
        cdef   bytes elt_name
        cdef   char* value_b
        cdef   bytes value_bytes
                
        if value is None :
            value_b = <char*>OBIStr_NA
        else :
            if self.name == b'sample' or self.name == b'SAMPLE':
                if type(value) == int:
                    value = str(value) # force sample ids to be str
            value_bytes = tobytes(value)
            value_b = <char*>value_bytes
                
        if type(elt_id) == int :
            if obi_set_str_with_elt_idx_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_id, <char*>value_b) < 0:
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")
        else :
            elt_name = tobytes(elt_id)
            if obi_set_str_with_elt_name_and_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, elt_name, <char*>value_b) < 0:
                obi_errno_to_exception(line_nb=line_nb, elt_id=elt_id, error_message="Problem setting a value in a column")


cdef class Column_tuples_str(Column_idx):
    
    cpdef object get_line(self, index_t line_nb) :
        global obi_errno
        cdef   const char*     value
        cdef   bytes           value_in_result
        cdef   object          result
        cdef   int32_t         i
        cdef   const char*     array
        cdef   int32_t         value_length
        
        result = []
        
        array = <const char*>obi_get_array_with_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, &value_length)
        obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem getting a value from a column")
        
        if array == OBITuple_NA :
            return None
        
        i = 0
        # First value
        value_in_result = <bytes> array
        result.append(value_in_result)
        while i+1 < value_length :
            if array[i] != b'\0' :
                i+=1
            else :
                value = array+i+1
                value_in_result = <bytes> value
                result.append(value_in_result)
                i+=1
        
        return tuple(result)
    
    cpdef set_line(self, index_t line_nb, object value) :
        global           obi_errno
        cdef   char*     array
        cdef   int32_t   value_length
        cdef   int32_t   i
        cdef   object    elt
        cdef   bytes     elt_b
                
        value_length = 0
        if value is not None:
            for i in range(len(value)) :
                if value[i] is not None and value[i] != '' :
                    value_length = value_length + len(value[i]) + 1   # Total size of the array with the '\0'               
        if value is None or value_length == 0 :
            array = <char*>OBITuple_NA
        else:
            array = <char*>calloc(value_length, sizeof(char))
            if array == NULL:
                raise Exception("Problem allocating memory for an array to store a tuple")
            #raise RollbackException("Problem allocating memory for an array to store a tuple", self._view) # TODO can't import
            i = 0
            for elt in value :
                if elt is not None and elt != '':
                    if self.name == b'sample' or self.name == b'SAMPLE':
                        if type(elt) == int:
                            elt = str(elt) # force sample ids to be str
                    elt_b = tobytes(elt)
                    strcpy(array+i, <char*>elt_b)
                    i = i + len(elt_b) + 1

        if obi_set_array_with_col_p_in_view(self._view.pointer(), self.pointer(), line_nb, <char*> array, sizeof(char)*8, value_length) < 0 :
            obi_errno_to_exception(line_nb=line_nb, elt_id=None, error_message="Problem setting a value in a column")
        
        if array != <char*>OBITuple_NA:
            free(array)


def register_class():
    register_column_class(OBI_STR, False, False, Column_str, bytes)
    register_column_class(OBI_STR, True, False, Column_multi_elts_str, bytes)
    register_column_class(OBI_STR, False, True, Column_tuples_str, bytes)

